from attrs import define, field, validators


@define(kw_only=True)
class AllianceKey:
    # Natural Keys

    name: str = field(
        validator=[
            validators.instance_of(str),
            validators.matches_re(r"^[\w ]{1,255}$"),
        ],
    )


@define(kw_only=True)
class Alliance(AllianceKey):
    pass



from typing import ClassVar

import diffsync
from cattrs import global_converter, structure, unstructure

from illallangi.rdf import RDFClient
from illallangi.rdf.diffsyncmodels import Course


@global_converter.register_structure_hook
def trip_structure_hook(
    value: dict,
    type: type,  # noqa: A002, ARG001
) -> Course:
    return Course(
        **value,
    )


class EducationAdapter(diffsync.Adapter):
    def __init__(
        self,
        *args: list,
        **kwargs: dict,
    ) -> None:
        super().__init__()
        self.client = RDFClient(
            *args,
            **kwargs,
        )

    Course = Course

    top_level: ClassVar = [
        "Course",
    ]

    type = "rdf_education"

    def load(
        self,
        *args: list,
        **kwargs: dict,
    ) -> None:
        for obj in self.client.get_courses(
            *args,
            **kwargs,
        ):
            d = unstructure(
                obj,
            )
            o = structure(
                d,
                Course,
            )
            self.add(
                o,
            )



from collections.abc import Generator
from typing import Any

from illallangi.rdf.models import Course


class CourseMixin:
    def get_courses(
        self,
        *_args: list[Any],
        rdf_root: str,
        **_kwargs: dict[str, Any],
    ) -> Generator[Course, Any, list | None]:
        query = f"""
SELECT ?start ?finish ?label ?institution ?street ?locality ?region ?postal_code ?country ?open_location_code WHERE {{
    <{ rdf_root }> ip:attendedCourse ?attended_course .
    OPTIONAL {{ ?attended_course ip:startTime ?start }} .
    OPTIONAL {{ ?attended_course ip:endTime ?finish }} .
    ?attended_course rdfs:label ?label .
    ?attended_course ip:atInstitution ?at_institution .

    OPTIONAL {{ ?at_institution rdfs:label ?institution . }}

    OPTIONAL {{ ?at_institution v:Address ?address
        OPTIONAL {{ ?address v:street-address ?street }}
        OPTIONAL {{ ?address v:locality ?locality }}
        OPTIONAL {{ ?address v:region ?region }}
        OPTIONAL {{ ?address v:postal-code ?postal_code }}
        OPTIONAL {{ ?address v:country-name ?country }}
    }}
    OPTIONAL {{ ?at_institution ip:olc ?open_location_code }} .
}}
"""

        for course in self.query(
            query=query,
        ):
            yield Course(
                **course,
            )



from collections.abc import Generator
from typing import Any

from illallangi.rdf.models import Airport


class AirportsNotFoundError(Exception):
    def __init__(
        self,
        airport_iata: list[str],
    ) -> None:
        self.airport_iata = airport_iata
        msg = f"Airport(s) not found: {', '.join(airport_iata)}"
        super().__init__(msg)


class AirportMixin:
    def get_airports(
        self,
        *_args: list[Any],
        airport_iata: list[str] | None = None,
        **_kwargs: dict[str, Any],
    ) -> Generator[Airport, Any, list | None]:
        if airport_iata is None:
            return []
        airport_iata = [i.upper() for i in airport_iata]

        query = f"""
SELECT ?label ?iata ?icao WHERE {{
    VALUES (?value) {{ ( "{'" ) ( "'.join(airport_iata)}" ) }}
    ?href ip:airportIataCode ?value.
    ?href rdfs:label ?label .
    ?href ip:airportIataCode ?iata .
    ?href ip:airportIcaoCode ?icao .
    ?href a ic:airport .
}}
"""

        found = []

        for airport in self.query(
            query=query,
        ):
            found.append(airport["iata"])
            yield Airport(
                **airport,
            )

        not_found = [i for i in airport_iata if i not in found]
        if not_found:
            raise AirportsNotFoundError(not_found)
